//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include "city.h"
#include "army.h"
#include "hero.h"
#include "stacklist.h"
#include "playerlist.h"
#include "armysetlist.h"
#include "defs.h"
#include "counter.h"
#include "GameMap.h"

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl<<flush;}
#define debug(x)

City::City(PG_Point pos, string name, int gold)
    :Location(name, pos, 2), d_player(0), d_numbasic(2),
    d_numadv(0), d_production(-1),
    d_adv_prod(false), d_duration(-1), d_gold(gold),
    d_defense_level(1), d_burnt(false)
{
    //assign id
    d_id = fl_counter->getNextId();

    // Initialise armytypes
    for (int i = 0; i < 4; i++)
        d_basicprod[i] = -1; 

    for (int i = 0; i < 2; i++)
        d_advprod[i] = -1;
}

City::City(XML_Helper* helper)
    :Location(), d_numbasic(2), d_numadv(0)
{
    d_size = 2;
    
    //initialize the city
    helper->getData(d_id, "id");
    helper->getData(d_name, "name");

    int i;
    helper->getData(i, "x");
    d_pos.x = i;
    helper->getData(i, "y");
    d_pos.y = i;
    helper->getData(i, "owner");
    d_player = Playerlist::getInstance()->getPlayer(static_cast<unsigned int>(i));

    //get production types
    istringstream sbase, sadv;
    string s;

    helper->getData(s, "basic_prod");
    sbase.str(s);
    for (i = 0; i < 4; i++)
        sbase >>d_basicprod[i];

    helper->getData(s, "adv_prod");
    sadv.str(s);
    for (i = 0; i < 2; i++)
        sadv >> d_advprod[i];

    helper->getData(d_defense_level, "defense");
    // Set the number of productions according to the defense level (level 1 is
    // included in the default initialization)
    if (d_defense_level == 2)
    {
        d_numbasic = 3;
        d_numadv = 1;
    }
    if (d_defense_level == 3)
    {
        d_numbasic = 4;
        d_numadv = 2;
    }
    
    helper->getData(d_production, "production");
    helper->getData(d_adv_prod, "is_adv_prod");
    helper->getData(d_duration, "duration");
    helper->getData(d_gold, "gold");
    helper->getData(d_burnt, "burnt");

    //mark the positions on the map as being occupied by a city
    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
            GameMap::getInstance()->getTile(d_pos.x+i, d_pos.y+j)
                                  ->setBuilding(Maptile::CITY);
}

City::~City()
{
}

bool City::save(XML_Helper* helper)
{
    bool retval = true;

    stringstream sbase, sadv;
    for (int i = 0; i < 4; i++)
        sbase << d_basicprod[i] <<" ";
    for (int i = 0; i < 2; i++)
        sadv << d_advprod[i] <<" ";

    retval &= helper->openTag("city");
    retval &= helper->saveData("id", d_id);
    retval &= helper->saveData("name", d_name);
    retval &= helper->saveData("x", d_pos.x);
    retval &= helper->saveData("y", d_pos.y);
    retval &= helper->saveData("owner", d_player->getId());
    retval &= helper->saveData("basic_prod", sbase.str());
    retval &= helper->saveData("adv_prod", sadv.str());
    retval &= helper->saveData("defense", d_defense_level);
    retval &= helper->saveData("production", d_production);
    retval &= helper->saveData("is_adv_prod", d_adv_prod);
    retval &= helper->saveData("duration", d_duration);
    retval &= helper->saveData("gold", d_gold);
    retval &= helper->saveData("burnt", d_burnt);
    retval &= helper->closeTag();
    return retval;
}

void City::setProduction(int index, bool advanced)
{
    if (index == -1)
    {
        d_production = index;
        d_adv_prod = false;
        d_duration = -1;
        return;
    }

    // return on wrong data
    if ((!advanced && (index >= d_numbasic)) || (advanced && (index >= d_numadv))
        || (index >= 0 && getArmytype(index, advanced) == -1))
        return;

    d_production = index;
    d_adv_prod = advanced;
    Armyset_army* a = getArmy(index, advanced);

    // set the duration to produce this armytype
    if (a)
        d_duration = a->getProduction(); 
}

bool City::raiseDefense()
{
    if (d_defense_level == 3)
    {
        // no further progress possible
        return false;
    }

    // increase the defense level, the income and the possible production
    d_defense_level++;
    d_gold = static_cast<int>(d_gold*1.5);
    d_numbasic++;
    d_numadv++;
    
    return true;
}

bool City::reduceDefense()
{
    if (d_defense_level == 1)
    {
        // no further reduction possible
        return false;
    }

    // the same as raiseDefense, but the other way round
    d_defense_level--;
    d_gold = static_cast<int>(d_gold*0.66);
    d_numbasic--;
    d_numadv--;

    // remove obsolete productions
    removeBasicProd(d_numbasic);
    removeAdvancedProd(d_numadv);
    
    return true;
}

bool City::addBasicProd(int index, int armytype)
{
    Armyset* as = Armysetlist::getInstance()->getArmyset("default");
    
    if ((index >= d_numbasic) || (armytype >= as->getSize()))
        return false;

    if (index < 0)
    {
        // try to find an unoccupied production slot. If there is none, pick 
        // the slot with the highest index.
        for (int i = 0; i < d_numbasic; i++)
            if (d_basicprod[i] == -1)
            {
                index = i;
                break;
            }

        if (index < 0)
        {
            index = d_numbasic - 1;
        }
    }
    
    removeBasicProd(index);
    d_basicprod[index] = armytype;
    return true;
}

void City::removeBasicProd(int index)
{
    if ((index < 0) || (index > 3))
        return;

    d_basicprod[index] = -1;

    if (!d_adv_prod && (d_production == index))
        setProduction(-1, false);
}

bool City::addAdvancedProd(int index, int armytype)
{
    if ((d_numadv == 0) || (index >= d_numadv) 
        || (armytype >= d_player->getArmyset()->getSize()))
        return false;

    if (index < 0)
    {
        // try to find an unoccupied production slot. If this doesn't work, 
        // overwrite the one with the highest index.
        for (int i = 0; i < d_numadv; i++)
            if (d_advprod[i] == -1)
            {
                index = i;
                break;
            }

        if (index < 0)
        {
            index = d_numadv - 1;
        }
    }

    removeAdvancedProd(index);
    d_advprod[index] = armytype;
    return true;
}

void City::removeAdvancedProd(int index)
{
    if ((index < 0) || (index > 1))
        return;

    d_advprod[index] = -1;
    if (d_adv_prod && (d_production == index))
        setProduction(-1, false);
}
            
void City::conquer(Player* newowner)
{
    d_player = newowner;

    // remove all advanced production
    for (int i = 0; i < 2; i++)
        removeAdvancedProd(i);
}

void City::setRandomArmytypes()
{
    //always set the lowest armytype
    addBasicProd(0, 0);

    //add another armytype if the dice are lucky
    int random_nr = rand() % 3;
    if (random_nr == 1) 
        addBasicProd(1, 1 + rand() % 4);
}

void City::produceLargestArmy()
{
    debug("produceLargestArmy()");

    // We assume that the largest army is the one with the highest index
    // with advanced productions ranking higher than normal ones.
    // This is a possible assumption as this function is solely used to
    // populate the cities during map generation.

    Stack* stack = getFreeStack();
    if (stack)
    {
        for (int adv = 1; adv >= 0; adv--)
            for (int i = 3; i >= 0; i--)
                if (getArmytype(i, adv) != -1)
                {
                    int savep = d_production;
                    bool saveadv = d_adv_prod;
                
                    setProduction(i, adv);
                    produceArmy();
                    setProduction(savep, saveadv);
                    return;
                }
    }
}

void City::addHero(Hero* hero)
{
    Stack* stack = getFreeStack();

    // No stack found so create one
    if (!stack)
    {
        stack = new Stack(d_player, d_pos);
        d_player->addStack(stack);
    }

    // add hero to stack
    stack->push_front(hero);
}

void City::nextTurn()
{
    if (d_burnt)
        return;

    // check if an army should be produced
    if (d_production >= 0 && --d_duration == 0) 
    {
        produceArmy();
    }
}

bool City::hasProduction(int type, Armyset* set)
{
    if (set == d_player->getArmyset() || set == 0)
        for (int i = 0; i < d_numadv; i++)
            if (d_advprod[i] == type)
                return true;

    if (set == Armysetlist::getInstance()->getArmyset("default"))
        for (int i = 0; i < d_numbasic; i++)
            if (d_basicprod[i] == type)
                return true;

    return false;
}

int City::getArmytype(int slot, bool advanced) const
{
    if (slot < 0)
        return -1;
    
    if (advanced)
    {
        if (slot >= d_numadv)
            return -1;
        return d_advprod[slot];
    }

    // !advanced
    if (slot >= d_numbasic)
        return -1;
    return d_basicprod[slot];
}

Armyset_army* City::getArmy(int slot, bool advanced) const
{
    if (getArmytype(slot, advanced) == -1)
        return 0;

    Armyset* as;
    
    if (advanced)
    {
        as = d_player->getArmyset();
        return &(as->army[d_advprod[slot]]);
    }
    
    as = Armysetlist::getInstance()->getArmyset("default");
    return &(as->army[d_basicprod[slot]]);
}

bool City::isFriend(Player* player) const
{
    return (d_player == player);
}

Stack* City::getFreeStack()
{
    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
        {
            Stack* stack = Stacklist::getObjectAt(d_pos.x + j, d_pos.y+ i);

            // No stack found so create one
            if (stack == 0)
            {
                PG_Point temp;
                temp.x = d_pos.x + j;
                temp.y = d_pos.y + i;
                stack = new Stack(d_player, temp);
                d_player->addStack(stack);
                return stack;
            }
            else if (stack->size() < 8) return stack;
        }

    return 0;
}

void City::produceArmy()
{
    debug("produce_army()\n");

    // do not produce an army if the player has no gold.
    if ((d_player->getGold() < 0) || (d_production == -1))
        return;

    Stack* stack = getFreeStack();

    if (stack)
    {
        // add produced army to stack
        Armyset* as;
        int index;
        
        if (d_adv_prod)
        {
            as = d_player->getArmyset();
            index = d_advprod[d_production];
        }
        else
        {
            as = Armysetlist::getInstance()->getArmyset("default");
            index = d_basicprod[d_production];
        }
        
        stack->push_back(new Army(as, index, d_player));
    }
    
    // start producing next army of same type
    setProduction(d_production, d_adv_prod);
}

// End of file
